////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   :  -   Compound File
//  * URL       : http://alexander-bagel.blogspot.ru/2015/03/compound-file.html
//  * Unit Name : Poifs/PoifsWithRepair
//  * Purpose   :     RAW    .
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2015.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//

unit PoifsWithRepair;

interface

uses
  Windows, Classes, SysUtils;

type
  USHORT = Word;

  TPoifsFileHeader = packed record
    // .   (0 x E011CFD0, 0 x E11AB1A1)
    _abSig: array [0..7] of Byte;
    // Class ID.  WriteClassStg,  GetClassFile/ReadClassStg.  Excel   = 0
    _clid: TGUID;
    //    .
    _uMinorVersion: USHORT;
    //    Dll/
    _uDllVersion: USHORT;
    // 0 x FFFE ,   Intel 
    _uByteOrder: USHORT;
    //  .   9,     512  (2 ^ 9)
    _uSectorShift: USHORT;
    //  -.   6,     64  (2 ^ 6)
    _uMiniSectorShift: USHORT;
    // ,    0
    _usReserved: USHORT;
    // ,    0
    _ulReserved1: ULONG;
    // ,    0
    _ulReserved2: ULONG;
    //  ,    FAT.
    //   <7,   1,  ,   1   DIF .
    _csectFat: ULONG;
    //   ,    Property Set Storage
    // (  FAT Directory  Root Directory Entry)
    _sectDirStart: ULONG;
    //   .
    _signature: ULONG;
    //   -.  4096
    _ulMiniSectorCutoff: ULONG;
    //   -FAT.
    //  0  FFFFFFFE (-2),  - .
    _sectMiniFatStart: ULONG;
    //     -FAT. 0,  - 
    _csectMiniFat: ULONG;
    //    DIF .
    //   <7,  DIF      0 x FFFFFFFE (-2)
    _sectDifStart: ULONG;
    //    DIF .0,   <7
    _csectDif: ULONG;
    //   109 ,    FAT.
    //   <7,   ,    0  FFFFFFFF (-1).
    _sectFat: array [0..108] of ULONG;
  end;

const
  HEADER_SIZE       = 512;
  STGTY_INVALID     = 0;
  STGTY_STORAGE     = 1;
  STGTY_STREAM      = 2;
  STGTY_LOCKBYTES   = 3;
  STGTY_PROPERTY    = 4;
  STGTY_ROOT        = 5;

  FREESECT          = -1;
  ENDOFCHAIN        = -2;

(*
typedef enum tagDECOLOR {
  DE_RED       = 0,
  DE_BLACK      = 1,
  } DECOLOR;
*)

type
  TPoifsProperty = packed record // 128 length
    //  /
    Caption: array[0..31] of WChar;
    //  
    CaptionSize: Word;
    //   STGTY_
    PropertyType: Byte;
    //   ( TPoifsProperty    Red-Black-Tree)
    NodeColor: Byte; // 0 (red) or 1 (black)
    //     
    PreviousProp: Integer;
    //     
    NextProp: Integer;
    //     
    ChildProp: Integer;
    Reserved1: TGUID;
    UserFlags: DWORD;
    // 
    ATime: array [0..1] of Int64;
    //   FAT       
    StartBlock: Integer;
    //  
    Size: Integer;
    Reserved2: DWORD;
  end;

  PPoifsProperty = ^TPoifsProperty;

  TPoifsFatBlock = array[0..127] of Integer;
  TPoifsBlock = array[0..128 * 4 - 1] of Byte;
  TPoifsPropsBlock = array[0..3] of TPoifsProperty;

  TPFErrorFoundEvent = procedure(Sender: TObject; FatIndex, Value: Integer;
    var Repair: Boolean) of Object;

  TPoifsFile = class(TPersistent)
  private
    FStream: TStream; // Data
    FHeader: TPoifsFileHeader;
    FFat: array of Integer; // Actual composed BAT
    FMiniFat: array of Integer;
    FFatOffset: array of Integer;
    FPropsArray: array of TPoifsProperty;
    FOnErrorFound: TPFErrorFoundEvent;
    FOnRepairResult: TPFErrorFoundEvent;
    procedure InitHeader;
    procedure ComposeFAT;
    procedure ComposeMiniFat;
    function ReadPropsArray: Boolean;
    function GetProperties(Index: Integer): TPoifsProperty;
    function GetPropertiesCount: Integer;
    function GetBlock(Adress: Integer): TPoifsBlock;
    function GetBlockOffset(BlockIndex: Integer): Int64;
    procedure FixFatEntry(FatIndex, NewValue: Integer);
  public
    constructor Create(Stream: TStream);
    function LoadFromStream: Boolean;
    procedure SetUserFlag(Index, Value: Integer);
    property Properties[Index: Integer]: TPoifsProperty read GetProperties; default;
    property PropertiesCount: Integer read GetPropertiesCount;
    property Header: TPoifsFileHeader read FHeader;
    property OnErrorFound: TPFErrorFoundEvent read FOnErrorFound write FOnErrorFound;
    property OnRepairResult: TPFErrorFoundEvent read FOnRepairResult write FOnRepairResult;
  public
    procedure GetDataFromMiniStream(ChainStart: ULONG;
      NeedLength: DWORD; const Stream: TStream);
    procedure GetDataFromStream(ChainStart: ULONG;
      NeedLength: DWORD; const Stream: TStream);
  end;

implementation

uses Math;

{ TPoifsFile }

//
//      
// =============================================================================
function TPoifsFile.LoadFromStream: Boolean;
begin
  Result := True;
  FStream.Position := 0;
  try
    //    
    InitHeader;
    //  FAT      (   _ulMiniSectorCutoff)
    ComposeFAT;
    //   FAT      ( _ulMiniSectorCutoff)
    ComposeMiniFat;
    //       
    ReadPropsArray;
  except
    Result := False;
  end;
end;

//
//   
// =============================================================================
procedure TPoifsFile.InitHeader;
begin
  FStream.ReadBuffer(FHeader, SizeOf(TPoifsFileHeader));
  FHeader._uSectorShift := Round(IntPower(2, FHeader._uSectorShift));
  FHeader._uMiniSectorShift := Round(IntPower(2, FHeader._uMiniSectorShift));
end;

//
//   FAT 
// =============================================================================
procedure TPoifsFile.ComposeFAT;
var
  I, J, X, FatLength: Integer;
  FatBlock: TPoifsFatBlock;
  CurrentFat, Offset: Integer;
  XFat: array of Integer;
begin
  //  -  FAT (   128 )
  //   DIF ,  _csectFat  
  FatLength := FHeader._csectFat * 128;

  //      FAT
  SetLength(FFat, FatLength);
  
  //       
  SetLength(FFatOffset, FatLength);

  //   DIF ,  FAT     109 
  //     DIF 
  for I := 0 to IfThen(FHeader._csectDif > 0, 108, FHeader._csectFat - 1) do
  begin
    //  FAT    128 
    FatBlock := TPoifsFatBlock(GetBlock(FHeader._sectFat[I]));
    for J := 0 to 127 do
    begin
      FFat[I * 128 + J] := FatBlock[J];
      //        , 
      FFatOffset[I * 128 + J] := FStream.Position - SizeOf(TPoifsBlock);
    end;
  end;

  // ,   DIF 
  if FHeader._sectDifStart = 0 then Exit;

  //  ,      FAT  
  Offset := FHeader._sectDifStart;

  //   XFAT     FAT 
  SetLength(XFat, 128);

  //     FAT 
  CurrentFat := 13951; //109 * 128 - 1 BAT

  for X := 0 to FHeader._csectDif - 1 do
  begin

    //     ( _uSectorShift  )
    //    
    FStream.Position := GetBlockOffset(Offset);

    //    FAT 
    FStream.ReadBuffer(XFat[0], 128 * SizeOf(DWORD));

    //     
    //       
    //       
    for I := 0 to 126 do
    begin

      //     ,   ,
      //  FAT  
      if XFat[I] < 0 then Exit;

      //  FAT    128 
      FatBlock := TPoifsFatBlock(GetBlock(XFat[I]));

      for J := 0 to 127 do
      begin
        Inc(CurrentFat);
        FFat[CurrentFat] := FatBlock[J];
        FFatOffset[CurrentFat] := FStream.Position - SizeOf(TPoifsBlock);
      end;
    end;

    //      
    Offset := XFat[127];
  end;
end;

//
//   
// =============================================================================
procedure TPoifsFile.ComposeMiniFat;
var
  I, CurrChain: Integer;
  TmpPosition: int64;
begin
  //       
  CurrChain := FHeader._sectMiniFatStart;

  //     (-    128 )
  SetLength(FMiniFat, FHeader._csectMiniFat * 128);

  I := 0;
  while Integer(CurrChain) >= 0 do
  begin
    //   
    TmpPosition := GetBlockOffset(CurrChain);

    //  ,   
    if TmpPosition < 0 then Exit;
    //if TmpPosition > FStream.Size then Exit;

    FStream.Position := TmpPosition;

    //  
    FStream.ReadBuffer(FMiniFat[I], 512 {128 * SizeOf(DWORD)});
    Inc(I, 128);

    //      FAT
    CurrChain := FFat[CurrChain];
  end;
end;

//
//   Property Set Storage
// =============================================================================
function TPoifsFile.ReadPropsArray: Boolean;
var
  I, J, Len, LastGood: Integer;
  PropsBlock: TPoifsPropsBlock;
begin
  Result := True;

  //    
  Len := 0;

  //    ,    Property Set Storage
  J := FHeader._sectDirStart;

  LastGood := J;

  repeat

    if J = FREESECT then
    begin
      FixFatEntry(LastGood, ENDOFCHAIN);
      Break;
    end;

    //     4 
    Inc(Len, 4);
    SetLength(FPropsArray, Len);

    //    
    try
      PropsBlock := TPoifsPropsBlock(GetBlock(J));
    except
      FixFatEntry(LastGood, ENDOFCHAIN);
      Break;
    end;

    for I := 0 to 3 do
      FPropsArray[Len - 4 + I] := PropsBlock[I];

    LastGood := J;

    //      FAT
    J := FFat[J];

    if J < ENDOFCHAIN then
    begin
      FixFatEntry(LastGood, ENDOFCHAIN);
      Break;
    end;

  until J = ENDOFCHAIN;
end;

//
//   ,      MiniFAT
// =============================================================================
procedure TPoifsFile.GetDataFromMiniStream(ChainStart: ULONG;
  NeedLength: DWORD; const Stream: TStream);
var
  MiniStreamOffset: DWORD;
  RealMiniStreamSector, TmpChain: Integer;
begin
  Stream.Size := 0;
  while (Integer(ChainStart) >= 0) and (Stream.Size < NeedLength) do
  begin
    //        Ministream
    TmpChain := ChainStart;
    RealMiniStreamSector := Properties[0].StartBlock;
    while TmpChain >= 8 do
    begin
      Dec(TmpChain, 8);
      RealMiniStreamSector := FFat[RealMiniStreamSector];
    end;
    //   
    MiniStreamOffset := GetBlockOffset(RealMiniStreamSector);

    //     
    FStream.Position := MiniStreamOffset +
      (ChainStart mod 8) * FHeader._uMiniSectorShift;

    //      
    ChainStart := FMiniFat[ChainStart];

    //   
    Stream.CopyFrom(FStream, FHeader._uMiniSectorShift);
  end;
  //  
  if Stream.Size > NeedLength then
    Stream.Size := NeedLength;
end;

//
//   ,      FAT
// =============================================================================
procedure TPoifsFile.GetDataFromStream(ChainStart: ULONG;
  NeedLength: DWORD; const Stream: TStream);
begin
  Stream.Size := 0;
  while (Integer(ChainStart) >= 0) and (Stream.Size < NeedLength) do
  begin
    //     
    FStream.Position := GetBlockOffset(ChainStart);

    //     
    ChainStart := FFat[ChainStart];

    //   
    Stream.CopyFrom(FStream, FHeader._uSectorShift);
  end;
  //  
  if Stream.Size > NeedLength then
    Stream.Size := NeedLength;
end;

// =============================================================================
//    
// =============================================================================

constructor TPoifsFile.Create(Stream: TStream);
begin
  FStream := Stream;
end;

procedure TPoifsFile.FixFatEntry(FatIndex, NewValue: Integer);
var
  J, Offset: Integer;
begin
  //    FAT 
  J := FatIndex mod 128;
  Offset := FFatOffset[FatIndex] + J * 4;

  //      
  FStream.Position := Offset;
  FStream.WriteBuffer(NewValue, SizeOf(Integer));
end;

function TPoifsFile.GetProperties(Index: Integer): TPoifsProperty;
begin
  Result := FPropsArray[Index];
end;

function TPoifsFile.GetPropertiesCount: Integer;
begin
  Result := Length(FPropsArray);
end;

function TPoifsFile.GetBlock(Adress: Integer): TPoifsBlock;
var
  BlockOffset: Integer;
begin
  BlockOffset := GetBlockOffset(Adress);
  if BlockOffset < FStream.Size then
  begin
    FStream.Position := BlockOffset;
    FStream.ReadBuffer(Result, SizeOf(TPoifsBlock));
  end
  else
    raise Exception.Create('Wrong block offset at addres: ' + IntToStr(Adress));
end;

function TPoifsFile.GetBlockOffset(BlockIndex: Integer): Int64;
begin
  Result := HEADER_SIZE + FHeader._uSectorShift * BlockIndex;
end;

procedure TPoifsFile.SetUserFlag(Index, Value: Integer);
begin
  FPropsArray[Index].UserFlags := Value;
end;

end.
