////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : VectoredGraph
//  * Purpose   :   
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2015.
//  * Version   : 1.04
//  ****************************************************************************
//

unit VectoredGraph;

interface

uses
  SysUtils,
  Classes;

const
  Header: string[3] = 'CBW';
  Version: Byte = 1;

type
  EAbstractGraph = class(Exception);
  EAbstractGraphVector = class(Exception);
  TAbstractGraphNode = class;
  TAbstractGraph = class;
  TAbstractGraphNodeClass = class of TAbstractGraphNode;
  TAbstractGraphVectorClass = class of TAbstractGraphVector;

  //   
  //       
  //           :)
  TDestroyNotifyer = class
  private
    FNotify: TNotifyEvent;
  protected
    procedure DoNotify; virtual;
  public
    destructor Destroy; override;
    property OnNotify: TNotifyEvent read FNotify write FNotify;
  end;

  TVectorLoadScruct = record
    PrimaryNodeID, SlaveNodeID: Integer;
    Primary, Slave: TAbstractGraphNode;
  end;
  TOnLoadEvent = procedure(Sender: TObject;
    var LoadScruct: TVectorLoadScruct) of object;

  //      ()
  //       (   )
  TAbstractGraphVector = class(TDestroyNotifyer)
  private
    FPrimaryNode: TAbstractGraphNode;
    FSlaveNode: TAbstractGraphNode;
    FID: Integer;
  private
    constructor Create(Index: Integer); overload;
  public
    constructor Create(Index: Integer;
      Primary, Slave: TAbstractGraphNode); overload; virtual;
    property ID: Integer read FID;
    property PrimaryNode: TAbstractGraphNode read FPrimaryNode;
    property SlaveNode: TAbstractGraphNode read FSlaveNode;
  end;

  //   ,
  //    ,   
  TAbstractGraphNode = class(TDestroyNotifyer)
  private
    FID: Integer;
    FGraphVectors: TList;
    FFreeVectorsOnDestroy: Boolean;
    function GetCount(const Index: Integer): Integer;
  protected
    property FreeVectorsOnDestroy: Boolean read FFreeVectorsOnDestroy write
      FFreeVectorsOnDestroy;
  public
    constructor Create(ID: Integer); virtual;
    destructor Destroy; override;
    procedure AddVector(Vector: TAbstractGraphVector);
    procedure DelVector(Vector: TAbstractGraphVector);
    function GetVector(const Index: Integer): TAbstractGraphVector;
    property ID: Integer read FID;
    property VectorCount: Integer index 0 read GetCount;
    property OutgouingVectorCount: Integer index 1 read GetCount;
    property IncomingVectorCount: Integer index 2 read GetCount;
  end;

  //  ,
  //     
  TAbstractGraph = class
  private
    FGraphVectors: TList;
    FGraphNodes: TList;
    function GetCount(const Index: Integer): Integer;
    function GetIndexFromID(const ID: Integer): Integer;
  protected
    function GetVectorClass: TAbstractGraphVectorClass; virtual;
    function GetNodeClass: TAbstractGraphNodeClass; virtual;
    procedure OnNodeNotify(Sender: TObject); virtual;
    procedure OnVectorNotify(Sender: TObject); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function AddVector(Primary, Slave: TAbstractGraphNode):
      TAbstractGraphVector; virtual;
    function AddNode(const ID: Integer): TAbstractGraphNode; virtual;
    procedure Clear;
    function GetNode(const Index: Integer): TAbstractGraphNode; virtual;
    function GetNodeByID(const Index: Integer): TAbstractGraphNode; virtual;
    function GetVector(const Index: Integer): TAbstractGraphVector; virtual;
    property VectorsCount: Integer index 0 read GetCount;
    property NodesCount: Integer index 1 read GetCount;
  end;

implementation

{ TDestroyNotifyer }

destructor TDestroyNotifyer.Destroy;
begin
  DoNotify;
  inherited;
end;

procedure TDestroyNotifyer.DoNotify;
begin
  if Assigned(FNotify) then
    FNotify(Self);
end;

{ TAbstractGraphVector }

//      
// =============================================================================
constructor TAbstractGraphVector.Create(Index: Integer;
  Primary, Slave: TAbstractGraphNode);
begin
  FID := Index;
  FPrimaryNode := Primary;
  FSlaveNode := Slave;
end;

//         
// =============================================================================
constructor TAbstractGraphVector.Create(Index: Integer);
begin
  FID := Index;
end;

{ TAbstractGraphNode }

procedure TAbstractGraphNode.AddVector(Vector: TAbstractGraphVector);
begin
  FGraphVectors.Add(Vector);
end;

constructor TAbstractGraphNode.Create(ID: Integer);
begin
  FGraphVectors := TList.Create;
  FreeVectorsOnDestroy := True;
  FID := ID;
end;

procedure TAbstractGraphNode.DelVector(Vector: TAbstractGraphVector);
var
  Index: Integer;
begin
  Index := FGraphVectors.IndexOf(Vector);
  if Index >= 0 then
    FGraphVectors.Delete(Index);
end;

destructor TAbstractGraphNode.Destroy;
var
  I: Integer;
begin
  //    -  ,
  //      
  //    (        )
  if FreeVectorsOnDestroy then
    for I := VectorCount - 1 downto 0 do
      GetVector(I).Free;
  FGraphVectors.Free;
  inherited;
end;

function TAbstractGraphNode.GetCount(const Index: Integer): Integer;
var
  I: Integer;
begin
  case Index of
    0: Result := FGraphVectors.Count;
    1:
    begin
      Result := 0;
      for I := 0 to VectorCount - 1 do
        Inc(Result, Byte(GetVector(I).PrimaryNode.ID = ID));
    end;
    2:
    begin
      Result := 0;
      for I := 0 to VectorCount - 1 do
        Inc(Result, Byte(GetVector(I).SlaveNode.ID = ID));
    end;
  else
    Result := -1;
  end;
end;

function TAbstractGraphNode.GetVector(
  const Index: Integer): TAbstractGraphVector;
begin
  Result := TAbstractGraphVector(FGraphVectors.Items[Index]);
end;

{ TAbstractGraph }

//      
// =============================================================================
function TAbstractGraph.AddNode(const ID: Integer): TAbstractGraphNode;
begin
  Result := GetNodeClass.Create(ID);
  Result.OnNotify := OnNodeNotify;
  FGraphNodes.Insert(GetIndexFromID(ID), Result);
end;

//      
// =============================================================================
function TAbstractGraph.AddVector(Primary,
  Slave: TAbstractGraphNode): TAbstractGraphVector;
begin
  //  
  Result := GetVectorClass.Create(VectorsCount, Primary, Slave);
  //    ,       
  FGraphVectors.Add(Result);
  //    
  Primary.AddVector(Result);
  Slave.AddVector(Result);
  //   ,
  //       
  Result.OnNotify := OnVectorNotify;
end;

//     
// =============================================================================
procedure TAbstractGraph.Clear;
var
  I: Integer;
  DestroyNotifyer: TDestroyNotifyer;
begin
  //        
  // ..     
  //     -    
  for I := VectorsCount - 1 downto 0 do
  begin
    DestroyNotifyer := TDestroyNotifyer(FGraphVectors[I]);
    DestroyNotifyer.OnNotify := nil;
    DestroyNotifyer.Free;
  end;
  FGraphVectors.Clear;
  for I := NodesCount - 1 downto 0 do
  begin
    DestroyNotifyer := TDestroyNotifyer(FGraphNodes[I]);
    //    ,    
    //          
    TAbstractGraphNode(DestroyNotifyer).FreeVectorsOnDestroy := False;
    DestroyNotifyer.OnNotify := nil;
    DestroyNotifyer.Free;
  end;
  FGraphNodes.Clear;
end;

constructor TAbstractGraph.Create;
begin
  FGraphVectors := TList.Create;
  FGraphNodes := TList.Create;
end;

destructor TAbstractGraph.Destroy;
begin
  Clear;
  FGraphNodes.Free;
  FGraphVectors.Free;
  inherited;
end;

function TAbstractGraph.GetCount(const Index: Integer): Integer;
begin
  case Index of
    0: Result := FGraphVectors.Count;
    1: Result := FGraphNodes.Count;
  else
    Result := -1;
  end;
end;

//       
// =============================================================================
function TAbstractGraph.GetNode(const Index: Integer): TAbstractGraphNode;
begin
  Result := TAbstractGraphNode(FGraphNodes.Items[Index]);
end;

//      ID  
// =============================================================================
function TAbstractGraph.GetNodeByID(const Index: Integer): TAbstractGraphNode;
var
  I: Integer;
  TempNode: TAbstractGraphNode;
begin
  Result := nil;
  I := GetIndexFromID(Index);
  TempNode := GetNode(I);
  if TempNode.ID = Index then
    Result := TempNode;
end;

//     
//       
// =============================================================================
function TAbstractGraph.GetNodeClass: TAbstractGraphNodeClass;
begin
  Result := TAbstractGraphNode;
end;

//          ,
//    ,   
// =============================================================================
function TAbstractGraph.GetIndexFromID(const ID: Integer): Integer;
var
  FLeft, FRight, FCurrent: Cardinal;
begin
  if NodesCount = 0 then
  begin
    Result := 0;
    Exit;
  end;
  FLeft := 0;
  FRight := NodesCount - 1;
  FCurrent := (FRight + FLeft) div 2;
  if GetNode(FLeft).ID > ID then
  begin
    Result := 0;
    Exit;
  end;
  if GetNode(FRight).ID < ID then
  begin
    Result := FRight + 1;
    Exit;
  end;
  repeat
    if GetNode(FCurrent).ID = ID then
    begin
      Result := FCurrent;
      Exit;
    end;
    if GetNode(FCurrent).ID < ID then
      FLeft := FCurrent
    else
      FRight := FCurrent;
    FCurrent := (FRight + FLeft) div 2;
  until FLeft = FCurrent;
  if GetNode(FCurrent).ID < ID then
    Inc(FCurrent);
  Result := FCurrent;
end;

function TAbstractGraph.GetVector(const Index: Integer): TAbstractGraphVector;
begin
  Result := TAbstractGraphVector(FGraphVectors.Items[Index]);
end;

//     
//       
// =============================================================================
function TAbstractGraph.GetVectorClass: TAbstractGraphVectorClass;
begin
  Result := TAbstractGraphVector;
end;

//       
// =============================================================================
procedure TAbstractGraph.OnNodeNotify(Sender: TObject);
var
  Index: Integer;
begin
  //     
  Index := FGraphNodes.IndexOf(TAbstractGraphNode(Sender));
  if Index >= 0 then
    FGraphNodes.Delete(Index);
end;

//       
// =============================================================================
procedure TAbstractGraph.OnVectorNotify(Sender: TObject);
var
  Vector: TAbstractGraphVector;
  Index: Integer;
begin
  Vector := TAbstractGraphVector(Sender);
  //     
  Vector.PrimaryNode.DelVector(Vector);
  Vector.SlaveNode.DelVector(Vector);
  //     
  Index := FGraphVectors.IndexOf(Vector);
  if Index >= 0 then
    FGraphVectors.Delete(Index);
end;

end.
