////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : FWStorage
//  * Purpose   :    IStorage/IStream
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2015.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//  !!!
//      ,        .
//    -        (  )
//        , ..      .
//

unit FWStorage;

interface

uses
  Windows,
  Classes,
  ActiveX,
  AxCtrls,
  SysUtils,
  ShlObj;

type
  TStatStgEx = record
    pacsName: string[30];
    dwType: Longint;
  end;
  TStatStgArray = array of TStatStgEx;
  
  TFWStorageEnum = packed record
    Count: DWORD;
    ElementEnum: TStatStgArray;
  end;

  EStorageException = class(Exception);
  TFWCustomStorageClass = class of TFWCustomStorage;
  TFWStorageCursorsClass = class of TFWStorageCursors;
  TFWStorageCursorClass = class of TFWStorageCursor;
  
  TFWStorageCursors = class;
  TFWCustomStorage = class;
  TFWStorage = class;

  TFWStorageCursor = class(TCollectionItem)
  private
    FLockCount: Integer;
    FID: ShortString;
    FStorageInterface: IStorage;
    FSubStorage: TFWStorageCursors;
    FMode: Longint;
    FOwner: TFWCustomStorage;
    FParent: TFWStorageCursor;
    FEnumerate: TFWStorageEnum;
    function GetPath: string;
    function GetLock: Boolean;
    function GetLockCount: Integer;
    procedure Lock;
    procedure UnLock;
    property Locked: Boolean read GetLock;
  protected
    function InternalOpen(const AName: AnsiString): HRESULT; virtual;
    function InternalCreate(const AName: AnsiString): HRESULT; virtual;
  protected
    procedure SetParam(const AOwner: TFWCustomStorage;
      AParent: TFWStorageCursor);
    property Owner: TFWCustomStorage read FOwner;
    property Parent: TFWStorageCursor read FParent;
  protected
    procedure Garbage;
    function GetObjectIndex(const AName: AnsiString; AType: Longint): Integer; virtual;
    procedure DeleteFromEnum(const AName: AnsiString; const AType: Longint); virtual;
    function GetCursorsClass: TFWStorageCursorsClass; virtual;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    function Backward(var ACursor: TFWStorageCursor): Boolean; virtual;
    function CreateStream(const AName: AnsiString): HRESULT; virtual;
    function CreateStorage(const AName: AnsiString;
      var ACursor: TFWStorageCursor): HRESULT;
    function Copy(const AName: AnsiString;
      const ACursor: TFWStorageCursor): HRESULT; virtual;
    function DeleteStream(const AName: AnsiString): HRESULT; virtual;
    function DeleteStorage(const AName: AnsiString): HRESULT; virtual;
    function Enumerate(var AData: TFWStorageEnum): HRESULT; virtual;
    function FlushBuffer: HRESULT; virtual;
    function MoveTo(const AName: AnsiString;
      const ACursor: TFWStorageCursor): HRESULT; virtual;
    function OpenStorage(const AName: AnsiString;
      var ACursor: TFWStorageCursor; Verify: Boolean = True): HRESULT;
    function ReadStream(const AName: AnsiString;
      AStream: TStream; Verify: Boolean = True): HRESULT; virtual;
    function Rename(const AName, ANewName: AnsiString): HRESULT; virtual;
    function WriteStream(const AName: AnsiString; AStream: TStream): HRESULT; virtual;
    procedure Release; reintroduce; virtual;
    function IsRoot: Boolean;
    function GetName: string; virtual;
    property Path: string read GetPath;
    property Storages: TFWStorageCursors read FSubStorage;
  end;

  TFWStorageCursors = class(TCollection)
  private
    function GetItem(Index: Integer): TFWStorageCursor;
    procedure SetItem(Index: Integer; const Value: TFWStorageCursor);
    function Add: TFWStorageCursor;
  protected
    function FindStorageByName(
      const AName: AnsiString; var Index: Cardinal): Boolean;
    function GetCursorClass: TFWStorageCursorClass; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    property Items[Index: Integer]: TFWStorageCursor read GetItem write SetItem; default;
  end;

  TFWCustomStorage = class(TPersistent)
  private
    FRootStorage: TFWStorageCursor;
    FMode: Longint;
    FFileName: string;
    FCanCreate: Boolean;
  protected
    function GetActive: Boolean; virtual;
    function OpenFileEx(const AFileName: string;
      const CanCreate: Boolean; const AMode: LongInt;
      var ACursor: TFWStorageCursor): HRESULT; virtual;
    procedure Garbage;
    class function ClassSupport(const AFileName: string; Change: Boolean): HRESULT;
    function GetCursorClass: TFWStorageCursorClass; virtual;
    function CreateCursor: TFWStorageCursor;
  public
    destructor Destroy; override;
    function ForceStorage(APath: string; var ACursor: TFWStorageCursor): HRESULT;
    function OpenFileReadOnly(const AFileName: string;
      var ACursor: TFWStorageCursor): HRESULT;
    function OpenFile(const AFileName: string;
      const CanCreate: Boolean; var ACursor: TFWStorageCursor): HRESULT;
    function OpenStorage(APath: string; var ACursor: TFWStorageCursor): HRESULT;
    procedure CloseFile; virtual;
    function ReConnect(AStorage: TFWStorage;
      var ACursor: TFWStorageCursor): HRESULT;
    class function Compress(const AFileName: string): HRESULT;
    class function IsStgValidBinaryFmt(const AFileName: string): HRESULT;
    property Active: Boolean read GetActive;
    property FileName: string read FFileName;
  end;

  TFWStorage = class(TFWCustomStorage);

implementation

uses
  Variants,
  ComObj;

{ TFWStorage }

//     Storage
// =============================================================================
procedure TFWCustomStorage.CloseFile;
begin
  if Assigned(FRootStorage) then
    FreeAndNil(FRootStorage);
end;

function TFWCustomStorage.CreateCursor: TFWStorageCursor;
begin
  if FRootStorage = nil then
  begin
    Result := nil;
    Exit;
  end;
  Result := FRootStorage;
end;

destructor TFWCustomStorage.Destroy;
begin
  CloseFile;
  inherited;
end;

function TFWCustomStorage.ReConnect(AStorage: TFWStorage;
  var ACursor: TFWStorageCursor): HRESULT;
begin
  AStorage.CloseFile;
  Result := AStorage.OpenFileEx(FFileName, FCanCreate, FMode, ACursor);
end;

//   ForceDirectory
// =============================================================================
function TFWCustomStorage.ForceStorage(APath: string;
  var ACursor: TFWStorageCursor): HRESULT;
var
  S: TStringList;
  I: Integer;
begin
  Result := S_FALSE;
  if APath = '' then
    raise EStorageException.Create(SysErrorMessage(ERROR_INVALID_PARAMETER));
  S := TStringList.Create;
  try
    ACursor := FRootStorage;
    if ACursor.GetName + '\' <> Copy(APath, 1, Length(ACursor.GetName) + 1) then
    begin
      Result := ERROR_PATH_NOT_FOUND;
      Exit;
    end;
    Delete(APath, 1, Length(ACursor.GetName) + 1);
    S.Text := StringReplace(APath, '\', sLineBreak, [rfReplaceAll]);
    for I := 0 to S.Count - 1 do
    begin
      if ACursor.GetObjectIndex(AnsiString(S.Strings[I]), STGTY_STORAGE) >= 0 then
        Result := ACursor.OpenStorage(AnsiString(S.Strings[I]), ACursor)
      else
        Result := ACursor.CreateStorage(AnsiString(S.Strings[I]), ACursor);
      if Result <> S_OK then
      begin
        FRootStorage.Garbage;
        Exit;
      end;
    end;
  finally
    S.Free;
  end;
end;      

procedure TFWCustomStorage.Garbage;
begin
  if FRootStorage <> nil then
    FRootStorage.Garbage;
end;

//    (   )
// =============================================================================
function TFWCustomStorage.OpenFile(const AFileName: string;
  const CanCreate: Boolean; var ACursor: TFWStorageCursor): HRESULT;
begin
  Result := OpenFileEx(AFileName, CanCreate,
    STGM_READWRITE or STGM_SHARE_EXCLUSIVE, ACursor);
end;

//    
// =============================================================================
function TFWCustomStorage.OpenFileEx(const AFileName: string;
  const CanCreate: Boolean; const AMode: LongInt;
  var ACursor: TFWStorageCursor): HRESULT;
begin
  FMode := AMode;
  FCanCreate := CanCreate;
  FFileName := AFileName;
  CloseFile;
  FRootStorage := GetCursorClass.Create(nil);
  if CanCreate then
  begin
    ACursor := FRootStorage;
    ACursor.SetParam(Self, nil);
    Result := StgCreateDocfile(StringToOleStr(AFileName), FMode or STGM_CREATE, 0,
      ACursor.FStorageInterface);
    ACursor.Lock;
  end
  else
  begin
    if not FileExists(AFileName) then
    begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    Result := StgIsStorageFile(StringToOleStr(AFileName));
    if Result <> S_OK then Exit;
    ACursor := FRootStorage;
    ACursor.SetParam(Self, nil);
    Result := StgOpenStorage(StringToOleStr(AFileName),
      nil, FMode, nil, 0, ACursor.FStorageInterface);
    ACursor.Lock;
  end;
end;

function TFWCustomStorage.OpenStorage(APath: string;
  var ACursor: TFWStorageCursor): HRESULT;
var
  S: TStringList;
  I: Integer;
begin
  Result := S_FALSE;
  if APath = '' then
    raise EStorageException.Create(SysErrorMessage(ERROR_INVALID_PARAMETER));
  S := TStringList.Create;
  try
    ACursor := FRootStorage;
    if APath[Length(APath)] <> '\' then APath := APath + '\';
    if ACursor.GetName + '\' <> Copy(APath, 1, Length(ACursor.GetName) + 1) then
    begin
      Result := ERROR_PATH_NOT_FOUND;
      Exit;
    end;
    Delete(APath, 1, Length(ACursor.GetName) + 1);
    S.Text := StringReplace(APath, '\', sLineBreak, [rfReplaceAll]);
    for I := 0 to S.Count - 1 do
      if ACursor.GetObjectIndex(AnsiString(S.Strings[I]), STGTY_STORAGE) >= 0 then
      begin
        if ACursor.OpenStorage(AnsiString(S.Strings[I]), ACursor) <> S_OK then Exit
      end
      else
        raise EStorageException.Create(SysErrorMessage(ERROR_PATH_NOT_FOUND));
    Result := S_OK;
  finally
    S.Free;
  end;
end;

//      
// =============================================================================
function TFWCustomStorage.OpenFileReadOnly(
  const AFileName: string; var ACursor: TFWStorageCursor): HRESULT;
begin
  Result := OpenFileEx(AFileName, False, STGM_READ or STGM_SHARE_EXCLUSIVE,
    ACursor);
end;

{ TFWStorageCursors }

function TFWStorageCursors.Add: TFWStorageCursor;
begin
  Result := TFWStorageCursor(inherited Add);
end;

constructor TFWStorageCursors.Create;
begin
  inherited Create(GetCursorClass);
end;

destructor TFWStorageCursors.Destroy;
begin
  inherited;
end;

//    Storage  
// =============================================================================
function TFWStorageCursors.FindStorageByName(
  const AName: AnsiString; var Index: Cardinal): Boolean;
var
  I: Cardinal;
  wName: string;
begin
  Result := False;
  wName := string(AName);
  Index := 0;
  if Count = 0 then Exit;
  for I := 0 to Count - 1 do
    if Items[I].GetName = wName then
    begin
      Result := True;
      Index := I;
      Break;
    end;
end;

function TFWStorageCursors.GetCursorClass: TFWStorageCursorClass;
begin
  Result := TFWStorageCursor;
end;

function TFWStorageCursors.GetItem(Index: Integer): TFWStorageCursor;
begin
  Result := TFWStorageCursor(inherited GetItem(Index));
end;

procedure TFWStorageCursors.SetItem(Index: Integer;
  const Value: TFWStorageCursor);
begin
  inherited SetItem(Index, Value);
end;

{ TFWStorageCursor }

//
// =============================================================================
function TFWStorageCursor.Backward(var ACursor: TFWStorageCursor): Boolean;
begin
  Result := FParent <> nil;
  if Result then
  begin
    UnLock;
    ACursor := FParent;
    ACursor.Lock;
    if not Locked then
      Self.Free;
  end;
end;

function TFWStorageCursor.Copy(const AName: AnsiString;
  const ACursor: TFWStorageCursor): HRESULT;
var
  I: Integer;
begin
  if ACursor = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_INVALIDARG)));
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  Result := FStorageInterface.MoveElementTo(StringToOleStr(AName),
    ACursor.FStorageInterface, StringToOleStr(AName), STGMOVE_COPY);

  if Result <> S_OK then Exit;
  if ACursor.FEnumerate.Count = MAXDWORD then Exit;
  for I := 0 to FEnumerate.Count - 1 do
    if AnsiCompareStr(FEnumerate.ElementEnum[I].pacsName, AName) = 0 then
    begin
      Inc(ACursor.FEnumerate.Count);
      SetLength(ACursor.FEnumerate.ElementEnum, ACursor.FEnumerate.Count);
      ACursor.FEnumerate.ElementEnum[ACursor.FEnumerate.Count - 1].pacsName := AName;
      ACursor.FEnumerate.ElementEnum[ACursor.FEnumerate.Count - 1].dwType :=
        FEnumerate.ElementEnum[I].dwType;
      Break;
    end;
end;

constructor TFWStorageCursor.Create(Collection: TCollection);
var
  ID: TGUID;
begin
  inherited;
  FEnumerate.Count := MAXDWORD;

  FSubStorage := GetCursorsClass.Create;
  FLockCount := 0;
  if CreateGUID(ID) <> S_OK then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  FID := ShortString(GUIDToString(ID));
end;

function TFWStorageCursor.CreateStorage(const AName: AnsiString;
  var ACursor: TFWStorageCursor): HRESULT;
var
  Index: Cardinal;
begin
  if FSubStorage.FindStorageByName(AName, Index) then
  begin
    Result := ERROR_ALREADY_EXISTS;
    //    -   ...
    OpenStorage(AName, ACursor);
    Exit;
  end;

  ACursor := FSubStorage.Add;
  ACursor.SetParam(FOwner, Self);
  Result := ACursor.InternalCreate(AName);

  if Result = STG_E_FILEALREADYEXISTS then
  begin
    ACursor.Free;
    Result := OpenStorage(AName, ACursor);
    Exit;
  end;
  
  ACursor.Lock;
  if FEnumerate.Count = MAXDWORD then Exit;
  if Result = S_OK then
  begin
    Inc(FEnumerate.Count);
    SetLength(FEnumerate.ElementEnum, FEnumerate.Count);
    FEnumerate.ElementEnum[FEnumerate.Count - 1].pacsName := AName;
    FEnumerate.ElementEnum[FEnumerate.Count - 1].dwType := STGTY_STORAGE;
  end;
end;

function TFWStorageCursor.CreateStream(const AName: AnsiString): HRESULT;
var
  TmpStream:IStream;
begin
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  Result := FStorageInterface.CreateStream(StringToOleStr(AName),
    FMode, 0, 0, TmpStream);
  TmpStream := nil;
  if FEnumerate.Count = MAXDWORD then Exit;
  if Result = S_OK then
  begin
    Inc(FEnumerate.Count);
    SetLength(FEnumerate.ElementEnum, FEnumerate.Count);
    FEnumerate.ElementEnum[FEnumerate.Count - 1].pacsName := AName;
    FEnumerate.ElementEnum[FEnumerate.Count - 1].dwType := STGTY_STREAM;
  end;
end;

procedure TFWStorageCursor.DeleteFromEnum(const AName: AnsiString;
  const AType: Integer);
var
  I, Index: Integer;
begin
  Index := GetObjectIndex(AName, AType);
  if Index >= 0 then
  begin
    if FEnumerate.Count = MAXDWORD then Exit;
    for I := Integer(Index) - 1 to Integer(FEnumerate.Count) - 2 do
      FEnumerate.ElementEnum[I] := FEnumerate.ElementEnum[I + 1];
    Dec(FEnumerate.Count);
    SetLength(FEnumerate.ElementEnum, FEnumerate.Count);
  end;
end;

function TFWStorageCursor.DeleteStorage(const AName: AnsiString): HRESULT;
var
  Index: Cardinal;
begin
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  if FSubStorage.FindStorageByName(AName, Index) then
  begin
    FSubStorage.Delete(Index);
    DeleteFromEnum(AName, STGTY_STORAGE);
  end;
  Result := FStorageInterface.DestroyElement(StringToOleStr(AName));
end;

function TFWStorageCursor.DeleteStream(const AName: AnsiString): HRESULT;
begin
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  Result := FStorageInterface.DestroyElement(StringToOleStr(AName));
  if Result = S_OK then
    DeleteFromEnum(AName, STGTY_STREAM);
end;

destructor TFWStorageCursor.Destroy;
begin
  FStorageInterface := nil;
  FreeAndNil(FSubStorage);
  inherited;
end;

function TFWStorageCursor.Enumerate(var AData: TFWStorageEnum): HRESULT;
var
  Enum: IEnumStatStg;
  TmpElement: TStatStg;
  ShellMalloc: IMalloc;
  Fetched: Int64;
  I: Integer;
begin
  if FEnumerate.Count <> MAXDWORD then
  begin
    AData.Count := FEnumerate.Count;
    SetLength(AData.ElementEnum, AData.Count);
    if FEnumerate.Count > 0 then    
      for I := 0 to FEnumerate.Count - 1 do
      begin
        AData.ElementEnum[I].pacsName := FEnumerate.ElementEnum[I].pacsName;
        AData.ElementEnum[I].dwType := FEnumerate.ElementEnum[I].dwType;
      end;
    Result := S_OK;
    Exit;
  end;
  
  if (CoGetMalloc(1, ShellMalloc) <> S_OK) or (ShellMalloc = nil) then
    raise EStorageException.Create('CoGetMalloc failed.');
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));

  Result := FStorageInterface.EnumElements(0, nil, 0, Enum);

  if Result = S_OK then
  begin
    FEnumerate.Count := 0;
    ZeroMemory(@FEnumerate, SizeOf(FEnumerate));
    Fetched := 1;
    while Fetched > 0 do
      if Enum.Next(1, TmpElement, @Fetched) = S_OK then
        if ShellMalloc.DidAlloc(TmpElement.pwcsName) = 1 then
        begin
          Inc(FEnumerate.Count);
          SetLength(FEnumerate.ElementEnum, FEnumerate.Count);
          FEnumerate.ElementEnum[FEnumerate.Count - 1].pacsName :=
            AnsiString(TmpElement.pwcsName);
          ShellMalloc.Free(TmpElement.pwcsName);
          FEnumerate.ElementEnum[FEnumerate.Count - 1].dwType := TmpElement.dwType;
        end;

    AData.Count := FEnumerate.Count;
    SetLength(AData.ElementEnum,  AData.Count);
    if FEnumerate.Count > 0 then
      for I := 0 to FEnumerate.Count - 1 do
      begin
        AData.ElementEnum[I].pacsName := FEnumerate.ElementEnum[I].pacsName;
        AData.ElementEnum[I].dwType := FEnumerate.ElementEnum[I].dwType;
      end;
  end;
  Enum := nil;
end;

function TFWStorageCursor.GetObjectIndex(const AName: AnsiString;
  AType: Longint): Integer;
var
  I: Integer;
  Data: TFWStorageEnum;
begin
  Result := -1;
  Data.Count := 0;
  if Enumerate(Data) = S_OK then
    if Data.Count > 0 then
      for I := 0 to Data.Count - 1 do
      begin
        if AnsiCompareText(Data.ElementEnum[I].pacsName, AName) = 0 then
        begin
          if Data.ElementEnum[I].dwType = AType then
            Result := I + 1;
          Break;
        end;
      end;
end;

function TFWStorageCursor.FlushBuffer: HRESULT;
begin
  Result := FStorageInterface.Commit(STGC_DEFAULT);
end;

procedure TFWStorageCursor.Garbage;
var
  I: Integer;
begin
  if FSubStorage = nil then Exit;
  if FSubStorage.Count > 0 then
    for I := 0 to FSubStorage.Count - 1 do
    begin
      FSubStorage.Items[I].Garbage;
      if not FSubStorage.Items[I].Locked then
        FSubStorage.Items[I].Free;
    end;
end;

function TFWStorageCursor.GetCursorsClass: TFWStorageCursorsClass;
begin
  Result := TFWStorageCursors;
end;

function TFWStorageCursor.GetLock: Boolean;
var
  I: Integer;
begin
  Result := FLockCount > 0;
  if not Result then
    if FSubStorage.Count > 0 then
      for I := 0 to FSubStorage.Count - 1 do
        if FSubStorage.Items[I].Locked then
        begin
          Result := True;
          Break;
        end;
end;

function TFWStorageCursor.GetLockCount: Integer;
var
  I: Integer;
begin
  Result := FLockCount;
  if FSubStorage.Count > 0 then
    for I := 0 to FSubStorage.Count - 1 do
      Inc(Result, FSubStorage.Items[I].GetLockCount);
end;

function TFWStorageCursor.GetName: string;
var
  Error: HRESULT;
  StatStg: TStatStg;
  ShellMalloc: IMalloc;
begin
  Result := '';
  if (CoGetMalloc(1, ShellMalloc) <> S_OK) or (ShellMalloc = nil) then
    raise EStorageException.Create('CoGetMalloc failed.');
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  Error := FStorageInterface.Stat(StatStg, STATFLAG_DEFAULT);
  if Error <> S_OK then
    raise EStorageException.Create(SysErrorMessage(Error));
  if ShellMalloc.DidAlloc(StatStg.pwcsName) = 1 then
  begin
    Result := string(StatStg.pwcsName);
    ShellMalloc.Free(StatStg.pwcsName);
  end;
end;

function TFWStorageCursor.GetPath: string;
begin
  if FParent = nil then
    Result := GetName
  else
    Result := FParent.GetPath + '\' + GetName;
end;

function TFWStorageCursor.InternalCreate(const AName: AnsiString): HRESULT;
begin
  Result := Parent.FStorageInterface.CreateStorage(StringToOleStr(AName),
     FMode, 0, 0, FStorageInterface);
end;

function TFWStorageCursor.InternalOpen(const AName: AnsiString): HRESULT;
begin
  Result := Parent.FStorageInterface.OpenStorage(StringToOleStr(AName),
    nil, FMode, nil, 0, FStorageInterface);
  if FStorageInterface = nil then
    Result := ERROR_PATH_NOT_FOUND;
end;

procedure TFWStorageCursor.Lock;
begin
  Inc(FLockCount);
end;

function TFWStorageCursor.MoveTo(const AName: AnsiString;
  const ACursor: TFWStorageCursor): HRESULT;
var
  Index: Cardinal;
  I: Integer;
begin
  if ACursor = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_INVALIDARG)));
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  if FSubStorage.FindStorageByName(AName, Index) then
    Result := E_ACCESSDENIED
  else
    Result := FStorageInterface.MoveElementTo(StringToOleStr(AName),
        ACursor.FStorageInterface, StringToOleStr(AName), STGMOVE_MOVE);

  if Result <> S_OK then Exit;
  if ACursor.FEnumerate.Count = MAXDWORD then Exit;
  for I := 0 to FEnumerate.Count - 1 do
    if AnsiCompareStr(FEnumerate.ElementEnum[I].pacsName, AName) = 0 then
    begin
      Inc(ACursor.FEnumerate.Count);
      SetLength(ACursor.FEnumerate.ElementEnum, ACursor.FEnumerate.Count);
      ACursor.FEnumerate.ElementEnum[ACursor.FEnumerate.Count - 1].pacsName := AName;
      ACursor.FEnumerate.ElementEnum[ACursor.FEnumerate.Count - 1].dwType :=
        FEnumerate.ElementEnum[I].dwType;
      DeleteFromEnum(AName, FEnumerate.ElementEnum[I].dwType);
      Break;
    end;
end;

function TFWStorageCursor.OpenStorage(const AName: AnsiString;
  var ACursor: TFWStorageCursor; Verify: Boolean = True): HRESULT;
var
  Index: Cardinal;
begin
  if fSubStorage.FindStorageByName(AName, Index) then
  begin
    Result := S_OK;
    ACursor := FSubStorage.Items[Index];
    ACursor.Lock;
    Exit;
  end;

  if Verify then
      if ACursor.GetObjectIndex(AName, STGTY_STORAGE) < 0 then
        raise EStorageException.Create(SysErrorMessage(ERROR_PATH_NOT_FOUND));

  if ACursor <> nil then
    if ACursor.FID = FID then UnLock;

  ACursor := FSubStorage.Add;
  ACursor.SetParam(FOwner, Self);

  Result := ACursor.InternalOpen(AName);
  if Result = ERROR_PATH_NOT_FOUND then
    raise EStorageException.Create(SysErrorMessage(ERROR_PATH_NOT_FOUND));
  ACursor.Lock;
end;

function TFWStorageCursor.ReadStream(const AName: AnsiString;
  AStream: TStream; Verify: Boolean = True): HRESULT;
var
  TmpStream:IStream;
  OS:TOleStream;
  Buff: array of Byte;
  I, Err: Integer;

  procedure DoReadAnything;
  begin
    //     ...
    SetLength(Buff, OS.Size);
    Err := 0;
    I := 0;
    OS.Position := 0;
    while I < OS.Size do
    try
      OS.Read(Buff[I - Err], 1);
      Inc(I);
    except
      Break;
    end;
    AStream.Write(Buff[0], I);
  end;

begin
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  if Verify then
  begin
    if GetObjectIndex(AName, STGTY_STREAM) >= 0 then
      Result := FStorageInterface.OpenStream(StringToOleStr(AName),
        nil, FMode, 0, TmpStream)
    else
      raise EStorageException.Create(SysErrorMessage(ERROR_PATH_NOT_FOUND));
  end
  else
    Result := FStorageInterface.OpenStream(StringToOleStr(AName),
      nil, FMode, 0, TmpStream);
  if Result = S_OK then
  begin
    OS := TOleStream.Create(TmpStream);
    try
      OS.Position := 0;
      try
        AStream.CopyFrom(OS, OS.Size);
      except
        on E: EOleException do
        begin
          Result := E.ErrorCode;
          DoReadAnything;
          Exit;
        end;
        on E: Exception do
        begin
          Result := S_FALSE;
          DoReadAnything;
        end;
      end;
      AStream.Position := 0;
    finally
      OS.Free;
      TmpStream := nil;
    end;
  end;
end;

procedure TFWStorageCursor.Release;
begin
  if FID = FOwner.FRootStorage.FID then Exit;
  Free;
  FOwner.Garbage;
end;

function TFWStorageCursor.Rename(const AName, ANewName: AnsiString): HRESULT;
var
  I: Integer;
begin
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  Garbage;
  Result := FStorageInterface.RenameElement(StringToOleStr(AName),
    StringToOleStr(ANewName));

  if FEnumerate.Count = MAXDWORD then Exit;
  for I := 0 to FEnumerate.Count - 1 do
    if AnsiCompareStr(FEnumerate.ElementEnum[I].pacsName, AName) = 0 then
    begin
      FEnumerate.ElementEnum[I].pacsName := ANewName;
      Break;
    end;
end;

procedure TFWStorageCursor.SetParam(const AOwner: TFWCustomStorage;
  AParent: TFWStorageCursor);
begin
  FMode := AOwner.FMode;
  FOwner := AOwner;
  FParent := AParent;
end;

procedure TFWStorageCursor.UnLock;
begin
  Dec(FLockCount);
end;

function TFWStorageCursor.WriteStream(const AName: AnsiString;
  AStream: TStream): HRESULT;
var
  TmpStream:IStream;
  OS:TOleStream;
begin
  if FStorageInterface = nil then
    raise EStorageException.Create(SysErrorMessage(Cardinal(E_UNEXPECTED)));
  if GetObjectIndex(AName, STGTY_STREAM) >= 0 then
    Result := FStorageInterface.OpenStream(StringToOleStr(AName),
      nil, FMode, 0, TmpStream)
  else
  begin
    //          
    //   ,    -    
    // ,  -     :)
    Result := FStorageInterface.CreateStream(StringToOleStr(AName),
      FMode, 0, 0, TmpStream);
    if Result <> S_OK then
    begin
      FStorageInterface.DestroyElement(StringToOleStr(AName));
      Result := FStorageInterface.CreateStream(StringToOleStr(AName),
        FMode, 0, 0, TmpStream);
    end;

    if FEnumerate.Count <> MAXDWORD then
      if Result = S_OK then
      begin
        Inc(FEnumerate.Count);
        SetLength(FEnumerate.ElementEnum, FEnumerate.Count);
        FEnumerate.ElementEnum[FEnumerate.Count - 1].pacsName := AName;
        FEnumerate.ElementEnum[FEnumerate.Count - 1].dwType := STGTY_STREAM;
      end;
      
  end;
  if Result = S_OK then
  begin
    OS := TOleStream.Create(TmpStream);
    try
      AStream.Position := 0;
      OS.CopyFrom(AStream, AStream.Size);
    finally
      OS.Free;
      TmpStream := nil;
    end;
  end;
end;

class function TFWCustomStorage.Compress(const AFileName: string): HRESULT;
begin
  Result := ClassSupport(AFileName, True);
end;

class function TFWCustomStorage.IsStgValidBinaryFmt(
  const AFileName: string): HRESULT;
begin
  Result := ClassSupport(AFileName, False);
end;

class function TFWCustomStorage.ClassSupport(const AFileName: string;
  Change: Boolean): HRESULT;
var
  Src, Dest: IStorage;
begin
  Result := S_FALSE;
  if not FileExists(AFileName) then Exit;
  Result := StgOpenStorage(StringToOleStr(AFileName), nil,
    STGM_READ or STGM_SHARE_EXCLUSIVE, nil, 0, Src);
  if Result = S_OK then
  try
    Result :=  StgCreateDocfile(StringToOleStr(string(AFileName) + '~'),
      STGM_READWRITE or STGM_SHARE_EXCLUSIVE or STGM_CREATE, 0, Dest);
    if Result = S_OK then
    try
      Result := Src.CopyTo(0, nil, nil, Dest);
    finally
      Dest := nil;
    end;
  finally
    Src := nil;
    if Change then
    begin
      if Result = S_OK then
        if not DeleteFile(AFileName) then
          Result := S_FALSE
        else
          RenameFile(AFileName + '~', AFileName);
    end
    else
    begin
      if Result = S_OK then
        if not DeleteFile(AFileName + '~') then
          Result := S_FALSE
    end;
  end;
end;

function TFWCustomStorage.GetActive: Boolean;
begin
  Result := FRootStorage <> nil;
  if Result then
    Result := FRootStorage.FStorageInterface <> nil;
end;

function TFWCustomStorage.GetCursorClass: TFWStorageCursorClass;
begin
  Result := TFWStorageCursor;
end;

function TFWStorageCursor.IsRoot: Boolean;
begin
  Result := FParent = nil;
end;

end.


